//
// NeoMagic.m
// NeoMagic Display Driver
//
// Copyright (c) 2004 David Lazaro Saz.
// All rights reserved.
//

#import "NeoMagic.h"
#include "NeoMagicModes.h"

#import <driverkit/generalFuncs.h>
#import <driverkit/i386/ioPorts.h>
#import <driverkit/i386/directDevice.h>
#import <driverkit/i386/IOPCIDirectDevice.h>
#import <driverkit/i386/IOPCIDeviceDescription.h>
#import <bsd/dev/ev_types.h>

#include <stdio.h>
#include <string.h>

#include "PCI.h"


/*
 * PCI Bus IDs
 */
#define PCI_VENDOR_NEOMAGIC 0x10C8

#define PCI_CHIP_NM2070 0x0001
#define PCI_CHIP_NM2090 0x0002
#define PCI_CHIP_NM2093 0x0003
#define PCI_CHIP_NM2097 0x0083
#define PCI_CHIP_NM2160 0x0004
#define PCI_CHIP_NM2200 0x0005
#define PCI_CHIP_NM2230 0x0025
#define PCI_CHIP_NM2360 0x0006
#define PCI_CHIP_NM2380 0x0016

/*
 * PCI Definitions
 */
#define PCI_MAP_MEMORY_ADDRESS_MASK 0xfffffff0


/* Default gamma precompensation table for color displays.
 * Gamma 2.2 LUT for P22 phosphor displays (Hitachi, NEC, generic VGA) */


static const unsigned char gamma16[] = {
      0,  74, 102, 123, 140, 155, 168, 180,
    192, 202, 212, 221, 230, 239, 247, 255
};

static const unsigned char gamma8[] = {
      0,  15,  22,  27,  31,  35,  39,  42,  45,  47,  50,  52,
     55,  57,  59,  61,  63,  65,  67,  69,  71,  73,  74,  76,
     78,  79,  81,  82,  84,  85,  87,  88,  90,  91,  93,  94,
     95,  97,  98,  99, 100, 102, 103, 104, 105, 107, 108, 109,
    110, 111, 112, 114, 115, 116, 117, 118, 119, 120, 121, 122,
    123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134,
    135, 136, 137, 138, 139, 140, 141, 141, 142, 143, 144, 145,
    146, 147, 148, 148, 149, 150, 151, 152, 153, 153, 154, 155, 
    156, 157, 158, 158, 159, 160, 161, 162, 162, 163, 164, 165,
    165, 166, 167, 168, 168, 169, 170, 171, 171, 172, 173, 174,
    174, 175, 176, 177, 177, 178, 179, 179, 180, 181, 182, 182,
    183, 184, 184, 185, 186, 186, 187, 188, 188, 189, 190, 190, 
    191, 192, 192, 193, 194, 194, 195, 196, 196, 197, 198, 198,
    199, 200, 200, 201, 201, 202, 203, 203, 204, 205, 205, 206, 
    206, 207, 208, 208, 209, 210, 210, 211, 211, 212, 213, 213,
    214, 214, 215, 216, 216, 217, 217, 218, 218, 219, 220, 220, 
    221, 221, 222, 222, 223, 224, 224, 225, 225, 226, 226, 227,
    228, 228, 229, 229, 230, 230, 231, 231, 232, 233, 233, 234, 
    234, 235, 235, 236, 236, 237, 237, 238, 238, 239, 240, 240,
    241, 241, 242, 242, 243, 243, 244, 244, 245, 245, 246, 246, 
    247, 247, 248, 248, 249, 249, 250, 250, 251, 251, 252, 252,
    253, 253, 254, 255, 
};

#define DACDelay(VGA)                                                        \
        do {                                                                 \
	    (void)inb((VGA)->PIOOffset + (VGA)->IOBase + VGA_IN_STAT_1_OFFSET); \
	    (void)inb((VGA)->PIOOffset + (VGA)->IOBase + VGA_IN_STAT_1_OFFSET); \
	} while (0)

static void SetGammaValue(VGADescription *VGA, const VGAMode *mode, int index,
                          unsigned int r, unsigned int g, unsigned int b,
                          int level)
{
    int shift = 0;
    
    switch (mode->depth) {
        case 15:
	default:
            shift = 2;
	    break;
    }
    
    VGA->writeDACWriteAddress(VGA, index);
    DACDelay(VGA);
    VGA->writeDACData(VGA, EV_SCALE_BRIGHTNESS(level, r) >> shift);
    DACDelay(VGA);
    VGA->writeDACData(VGA, EV_SCALE_BRIGHTNESS(level, g) >> shift);
    DACDelay(VGA);
    VGA->writeDACData(VGA, EV_SCALE_BRIGHTNESS(level, b) >> shift);
    DACDelay(VGA);
}


@interface NeoMagic (Private)

- (void)initVGA;
- (void)freeVGA;

- (void)lock;
- (void)unlock;

- (void)saveMode;
- (void)programShadowRegisters:(NeoMagicRegisters *)reg;
- (void)restoreModeWithRegisters:(NeoMagicRegisters *)reg
                    VGARegisters:(VGARegisters *)vgareg;

- (void)setGammaTable;
- (void)calcVCLKFromFreq:(long)freq;

@end


@implementation NeoMagic

+ (BOOL)probe:(id)deviceDescription
{
    unsigned char deviceNumber   = 0;
    unsigned char functionNumber = 0;
    unsigned char busNumber      = 0;
    
    unsigned long  vendorDeviceID = 0;
    unsigned short vendorID       = 0;
    unsigned short deviceID       = 0;
    char           idString[11];
    
    IOPCIConfigSpace  configSpace;
    unsigned int      numMemoryRanges;
    IORange          *oldRanges;
    IORange           newRanges[3];
	
    unsigned int      frameBufferSize = 0;
    
    IOReturn status = IO_R_SUCCESS;
    
    int i;
    
    IOConfigTable *configTable = [deviceDescription configTable];
    
    NeoMagic *neoMagic;
    
    // Check for the PCI bus.
    if (![IODirectDevice isPCIPresent]) {
        IOLog("%s: PCI Bus not found\n", [self name]);
        return NO;
    }
    
    // Caracterize the device.
    status = [deviceDescription getPCIdevice:&deviceNumber
      	                            function:&functionNumber
			                 bus:&busNumber];
    if (status == IO_R_SUCCESS) {
        IOLog("%s: PCI Device:%d Function:%d Bus:%d\n", [self name],
              deviceNumber, functionNumber, busNumber);
    } else {
        IOLog("%s: Could not get PCI config address\n", [self name]);
	return NO;
    }
    
    status = [IODirectDevice getPCIConfigData:&vendorDeviceID
                                   atRegister:0x00
		        withDeviceDescription:deviceDescription];
    if (status != IO_R_SUCCESS) {
        IOLog("%s: Could not read device configuration space.\n", [self name]);
	return NO;
    }
    vendorID = (unsigned short)vendorDeviceID;
    deviceID = (unsigned short)(vendorDeviceID >> 16);
    IOLog("%s: VendorID:%04x DeviceID:%04x\n", [self name],
          vendorID, deviceID);
    
    sprintf(idString, "%08lx", vendorDeviceID);
    if (strstr([configTable valueForStringKey:"Auto Detect IDs"], idString) ==
        NULL) {
	IOLog("%s: Vendor Device ID %08lx not found in instance table.\n",
	      [self name], vendorDeviceID);
	return NO;
    } else {
        IOLog("%s: NeoMagic ", [self name]);
        switch(deviceID) {
            case PCI_CHIP_NM2070:
	        IOLog("MagicGraph 128");
		frameBufferSize = 1024 * 1024;
		break;
            case PCI_CHIP_NM2090:
	        IOLog("MagicGraph 128V");
		frameBufferSize = 2048 * 1024;
		break;
            case PCI_CHIP_NM2093:
	        IOLog("MagicGraph 128ZV");
		frameBufferSize = 2048 * 1024;
		break;
            case PCI_CHIP_NM2097:
	        IOLog("MagicGraph 128ZV+");
		frameBufferSize = 2048 * 1024;
		break;
            case PCI_CHIP_NM2160:
	        IOLog("MagicGraph 128XD");
		frameBufferSize = 2048 * 1024;
		break;
            case PCI_CHIP_NM2200:
	        IOLog("MagicMedia 256AV");
		frameBufferSize = 4096 * 1024;
		break;
            case PCI_CHIP_NM2230:
	        IOLog("MagicMedia 256AV+");
		frameBufferSize = 4096 * 1024;
		break;
            case PCI_CHIP_NM2360:
	        IOLog("MagicMedia 256ZX");
		frameBufferSize = 4096 * 1024;
		break;
            case PCI_CHIP_NM2380:
	        IOLog("MagicMedia 256XL+");
		frameBufferSize = 8192 * 1024;
		break;
	}
	IOLog(" display adapter found.\n");
    }
        
    // Update the frame buffer memory ranges.
    status = [IODirectDevice getPCIConfigSpace:&configSpace
                         withDeviceDescription:deviceDescription];
    if (status != IO_R_SUCCESS) {
        IOLog("%s: Could not get configuration space.\n", [self name]);
	return NO;
    }
    numMemoryRanges = [deviceDescription numMemoryRanges];
    if (numMemoryRanges != 3) {
        IOLog("%s: Incorrect number of memory ranges: %d.\n", [self name],
	      numMemoryRanges);
	return NO;
    }
    oldRanges = [deviceDescription memoryRangeList];
    for (i = 0; i < numMemoryRanges; i++)
        newRanges[i] = oldRanges[i];
    newRanges[0].start = configSpace.BaseAddress[0] &
	                 PCI_MAP_MEMORY_ADDRESS_MASK;
    newRanges[0].size = frameBufferSize;
	
    status = [deviceDescription setMemoryRangeList:newRanges
                                               num:numMemoryRanges];
    if (status != IO_R_SUCCESS) {
        IOLog("%s: Could not set the memory range list.\n", [self name]);
	return NO;
    } else
        IOLog("%s: Frame buffer at 0x%08x-0x%08x.\n", [self name],
	      newRanges[0].start,
	      newRanges[0].start + newRanges[0].size);
    
    // Create and initialize the driver instance.
    neoMagic = [self alloc];
    return ([neoMagic initFromDeviceDescription:deviceDescription] ? YES : NO);
}


- (id)initFromDeviceDescription:(id)deviceDescription
{
    BOOL validModes[NeoMagicModeCount];
    int  mode;
    
    IORange      *memoryRangeList;
    vm_address_t  address;
    
    IODisplayInfo *displayInfo;
    
    unsigned long  vendorDeviceID = 0;
    unsigned short vendorID       = 0;
    unsigned short deviceID       = 0;
    char           idString[11];

    IOReturn status = IO_R_SUCCESS;
    
    unsigned char type;
    unsigned char dpy;
    unsigned char w;

    int i;
    
    if ([super initFromDeviceDescription:deviceDescription] == nil)
        return [super free];

    // Select a display mode.
    for (i = 0; i < NeoMagicModeCount; i++) {
        validModes[i] = YES;
    }
    mode = [self selectMode:NeoMagicModeTable
                      count:NeoMagicModeCount
		      valid:validModes];
    if (mode < 0) {
        IOLog("%s: Sorry, cannot use requested display mode.\n", [self name]);
	mode = 1;  /* Default mode */
    }
    displayMode = &NeoMagicModeTable[mode];
    vgaMode = &NeoMagicCRTCModeTable[mode];
    
    IOLog("Selected mode at index: %d\n", mode);

    // Map the frame buffer.
    memoryRangeList = [deviceDescription memoryRangeList];
    address = [self mapFrameBufferAtPhysicalAddress:memoryRangeList[0].start
                                             length:memoryRangeList[0].size];
    if (address == NULL) {
        IOLog("%s: Frame buffer memory mapping failed.\n", [self name]);
        return [super free];
    } else
        IOLog("%s: Frame buffer (0x%x-0x%x) mapped at 0x%x.\n", [self name],
	      memoryRangeList[0].start,
	      memoryRangeList[0].start + memoryRangeList[0].size,
	      address);

    displayInfo = [self displayInfo];
    *displayInfo = *displayMode;
    displayInfo->frameBuffer = (void *)address;
    
    /* Get the chipset model. */
    status = [IODirectDevice getPCIConfigData:&vendorDeviceID
                                   atRegister:0x00
		        withDeviceDescription:deviceDescription];
    if (status != IO_R_SUCCESS) {
        IOLog("%s: Could not read device configuration space.\n", [self name]);
	return NO;
    }
    vendorID = (unsigned short)vendorDeviceID;
    deviceID = (unsigned short)(vendorDeviceID >> 16);
    
    sprintf(idString, "%08lx", vendorDeviceID);
    if (strstr([[deviceDescription configTable]
                    valueForStringKey:"Auto Detect IDs"], idString) == NULL) {
	IOLog("%s: Vendor Device ID %08lx not found in instance table.\n",
	      [self name], vendorDeviceID);
	return NO;
    } else {
        switch(deviceID) {
            case PCI_CHIP_NM2070:
		chipset = NM2070;
		break;
            case PCI_CHIP_NM2090:
		chipset = NM2090;
		break;
            case PCI_CHIP_NM2093:
		chipset = NM2093;
		break;
            case PCI_CHIP_NM2097:
		chipset = NM2097;
		break;
            case PCI_CHIP_NM2160:
		chipset = NM2160;
		break;
            case PCI_CHIP_NM2200:
		chipset = NM2200;
		break;
            case PCI_CHIP_NM2230:
		chipset = NM2230;
		break;
            case PCI_CHIP_NM2360:
		chipset = NM2360;
		break;
            case PCI_CHIP_NM2380:
		chipset = NM2380;
		break;
	}
    }
    
    [self initVGA];

    /* Determine the display type. */
    VGA->writeGraphics(VGA, 0x09, 0x26);
    type = VGA->readGraphics(VGA, 0x21);
    dpy = VGA->readGraphics(VGA, 0x20);
    
    /* Determine the panel size. */
    w = VGA->readGraphics(VGA, 0x20);
    VGA->writeGraphics(VGA, 0x09, 0x00);
    switch ((w & 0x18) >> 3) {
        case 0x00:
	    panelWidth  = 640;
	    panelHeight = 480;
	    break;
	case 0x01:
	    panelWidth  = 800;
            panelHeight = (mode == LibrettoDisplayModeIndex) ? 480 : 600; //Assume Libretto panel if 800x480 mode selected
	    break;
	case 0x02:
	    panelWidth  = 1024;
	    panelHeight = 768;
	    break;
	default:
	    panelHeight  = 640;
	    panelWidth   = 480;
	    break;
    }
    IOLog("%s: Panel size is a %d x %d %s %s display. Mode: 0x%x.\n",
          [self name],
          panelWidth, panelHeight,
	  (type & 0x02) ? "color" : "monochrome",
	  (type & 0x10) ? "TFT" : "dual scan",
	  dpy);

    redTransferTable = greenTransferTable = blueTransferTable = 0;
    transferTableCount = 0;
    brightnessLevel = EV_SCREEN_MAX_BRIGHTNESS / 2;

    [self setDeviceKind:"Linear Framebuffer"];
    [self registerDevice];

    return self;
}


- (id)free
{
    [self freeVGA];

    return [super free];
}


- (void)enterLinearMode
{
    int hOffset;
    int vOffset;
    
    int clockMul;
    
    IODisplayInfo *displayInfo;
    
    clockMul = 1;
    
    [self saveMode];
    
    [self unlock];
    
    /* Initialize the generic VGA registers. */
    VGAEnterMode(VGA, vgaRegisters, vgaMode);
    
    /*
     * The default value assigned by VGAEnterMode is 0x41, but this
     * does not work for NeoMagic.
     */
    vgaRegisters->attribute[16] = 0x01;
    
    switch (vgaMode->depth) {
        case 15:
	    registers->extendedColorModeSelect = 0x12;
	    vgaRegisters->CRTC[0x13] = displayMode->width >> 2;
	    registers->externalCRTOffset = displayMode->width >> 10;
	    break;
	default:
	    break;
    }
    
    registers->externalCRTDisplayAddress = 0x10;
    
    /* Vertical Extension */
    registers->verticalExtension = (((vgaMode->vTotal - 2) & 0x400) >> 10) |
                                   (((vgaMode->vDisplay - 1) & 0x400) >> 9) |
				   (((vgaMode->vSyncStart) & 0x400) >> 8) |
				   (((vgaMode->vBlankStart - 1) & 0x400) >> 7);

    /* Fast write bursts on. */
    registers->systemInterfaceControl1 = 0x30;
    
    /* Enable linear addressing. */
    registers->systemInterfaceControl2 = 0x80;
    
    /* Enable only the internal display. */
    registers->panelDisplayControl1 = 0x02;
    
    /* If we are using a fixed mode tell the chip we are. */
    switch (vgaMode->hDisplay) {
        case 1280:
	    registers->panelDisplayControl1 |= 0x60;
	    break;
	case 1024:
	    registers->panelDisplayControl1 |= 0x40;
	    break;
	case 800:
	    registers->panelDisplayControl1 |= 0x20;
	case 640:
	default:
	    break;
    }
    
    /* Setup shadow register locking. */
    switch (registers->panelDisplayControl1 & 0x03) {
        case 0x01:
	    /* External CRT only mode. */
	    registers->generalLock = 0x00;
	    /* We need to program the VCLK for external display only mode. */
	    registers->programVCLK = 0x01;
	    break;
	case 0x02:
	    /* Internal LCD only mode. */
	case 0x03:
	    /* Simultaneous internal/external mode. */
	    registers->generalLock = 0x01;
	    registers->programVCLK = 0x00;
	    break;
    }
    
    /* Turn off stretching. */
    registers->panelDisplayControl2 = 0x00;
    registers->panelDisplayControl3 = 0x00;
    
    /* Turn on centering for the various modes. */
    registers->panelVerticalCenter1 = 0x00;
    registers->panelVerticalCenter2 = 0x00;
    registers->panelVerticalCenter3 = 0x00;
    registers->panelVerticalCenter4 = 0x00;
    registers->panelVerticalCenter5 = 0x00;
    registers->panelHorizontalCenter1 = 0x00;
    registers->panelHorizontalCenter2 = 0x00;
    registers->panelHorizontalCenter3 = 0x00;
    registers->panelHorizontalCenter4 = 0x00;
    registers->panelHorizontalCenter5 = 0x00;
    
    if (registers->panelDisplayControl1 & 0x02) {
        unsigned char doCenter = 0x00;
	
	hOffset = vOffset = 0;
	
	if (vgaMode->hDisplay == panelWidth) {
	    /*
	     * No centering required when the requested display width
	     * equals the panel width.
	     */
	} else {
	    int hDisplay;
	    
	    registers->panelDisplayControl3 |= 0x10;
	    hDisplay = vgaMode->hDisplay <<
	               ((vgaMode->vDisplay < 480) ? 1 : 0);
	    hOffset = ((panelWidth - hDisplay) >> 4) - 1;
	    if (vgaMode->vDisplay < 480)
	        hOffset >>= 1;
	    doCenter = 0x01;
	}
	if (vgaMode->vDisplay == panelHeight) {
	    /*
	     * No centering required when te requisted display height
	     * equals the panel height.
	     */
	} else {
	    int vDisplay;
	    
	    registers->panelDisplayControl2 |= 0x01;
	    vDisplay = vgaMode->vDisplay <<
	               ((vgaMode->flags & VGAModeDoubleScan) ? 1 : 0);
	    vOffset = ((panelHeight - vDisplay) >> 1) - 2;
	    doCenter = 0x01;
	}
	
	if (doCenter) {
	    switch (vgaMode->hDisplay) {
	        case 320:
		    registers->panelHorizontalCenter3 = hOffset;
		    registers->panelVerticalCenter3 = vOffset;
		    break;
		case 400:
		    registers->panelHorizontalCenter4 = hOffset;
		    registers->panelVerticalCenter1 = vOffset;
		    break;
		case 640:
		    registers->panelHorizontalCenter1 = hOffset;
		    registers->panelVerticalCenter3 = vOffset;
		    break;
		case 800:
		    registers->panelHorizontalCenter2 = hOffset;
		    switch (vgaMode->vDisplay) {
		        case 600:
		            registers->panelVerticalCenter4 = vOffset;
			    break;
			case 480:
			    registers->panelVerticalCenter3 = vOffset;
			    break;
		    }
		    break;
		case 1024:
		    registers->panelHorizontalCenter5 = hOffset;
		    registers->panelVerticalCenter5 = vOffset;
		    break;
		case 1280:
		default:
		    /* No centering in these modes. */
		    break;
	    }
	}
    }
    
    if (vgaMode->vDisplay < 480) {
        vgaRegisters->sequencer[1] |= 0x08;
        clockMul = 2;
    }
    
    registers->biosMode = vgaMode->biosMode;
    
    registers->useExtendedRegisters = 0x00;
    
    /* Calculate the VCLK. */
    [self calcVCLKFromFreq:vgaMode->clock * clockMul];
    
    /* Use VCLK3 as the clock will be programmed. */
    vgaRegisters->miscOut |= 0x0C;
    
    VGAProtect(VGA);
    
    [self restoreModeWithRegisters:registers VGARegisters:vgaRegisters];
    
    [self setGammaTable];
    
    /* Clear the screen to black. */
    displayInfo = [self displayInfo];
    memset(displayInfo->frameBuffer, 0,
           displayInfo->rowBytes * displayInfo->height);
    
    VGAUnprotect(VGA);
    //[self lock];
    
}


- (void)revertToVGAMode
{
    //[self unlock];
    VGAProtect(VGA);
    
    [self restoreModeWithRegisters:savedRegisters
                      VGARegisters:savedVGARegisters];

    VGAUnprotect(VGA);
    [self lock];
}

- setBrightness:(int)level token:(int)t
{
    if (level < EV_SCREEN_MIN_BRIGHTNESS || level > EV_SCREEN_MAX_BRIGHTNESS) {
	IOLog("%s: Invalid arg to setBrightness: %d\n", [self name], level);
	return nil;
    }

    brightnessLevel = level;

    [self setGammaTable];

    return self;
}

- (id)setTransferTable:(const unsigned int *)table count:(int)numEntries
{
    int k;
    
    IOBitsPerPixel bitsPerPixel;
    IOColorSpace   colorSpace;

    if (redTransferTable != 0)
        IOFree(redTransferTable, 3 * transferTableCount);

    transferTableCount = numEntries;

    redTransferTable   = IOMalloc(3 * numEntries);
    greenTransferTable = redTransferTable + numEntries;
    blueTransferTable  = greenTransferTable + numEntries;

    bitsPerPixel = [self displayInfo]->bitsPerPixel;
    colorSpace   = [self displayInfo]->colorSpace;

    if (bitsPerPixel == IO_8BitsPerPixel &&
        colorSpace   == IO_OneIsWhiteColorSpace) {
        for (k = 0; k < numEntries; k++) {
            redTransferTable[k] = greenTransferTable[k] =
            blueTransferTable[k] = table[k] & 0xFF;
        }
    } else if (colorSpace == IO_RGBColorSpace &&
               (bitsPerPixel == IO_8BitsPerPixel ||
                bitsPerPixel == IO_12BitsPerPixel ||
                bitsPerPixel == IO_15BitsPerPixel ||
                bitsPerPixel == IO_24BitsPerPixel)) {
        for (k = 0; k < numEntries; k++) {
            redTransferTable[k]   = (table[k] >> 24) & 0xFF;
            greenTransferTable[k] = (table[k] >> 16) & 0xFF;
            blueTransferTable[k]  = (table[k] >> 8) & 0xFF;
        }
    } else {
        IOLog("%s: Can't set transfer table; unknown colorSpace (%d)"
              " and pixel depth (%d)\n",
              [self name], colorSpace, bitsPerPixel);
        IOFree(redTransferTable, 3 * numEntries);
        redTransferTable = 0;
    }

    [self setGammaTable];

    return self;
}

@end


@implementation NeoMagic (Private)

- (void)initVGA
{
    VGA = IOMalloc(sizeof(VGADescription));
    
    vgaRegisters = IOMalloc(sizeof(VGARegisters));
    savedVGARegisters = IOMalloc(sizeof(VGARegisters));
    registers = IOMalloc(sizeof(NeoMagicRegisters));
    savedRegisters = IOMalloc(sizeof(NeoMagicRegisters));

    vgaRegisters->numCRTC = savedVGARegisters->numCRTC = VGA_NUM_CRTC;
    vgaRegisters->numSequencer = savedVGARegisters->numSequencer = VGA_NUM_SEQ;
    vgaRegisters->numGraphics = savedVGARegisters->numGraphics = VGA_NUM_GFX;
    vgaRegisters->numAttribute = savedVGARegisters->numAttribute = VGA_NUM_ATTR;

    VGARegistersAlloc(vgaRegisters);
    VGARegistersAlloc(savedVGARegisters);
    
    VGAGetDescription(VGA, vgaRegisters);
}

- (void)freeVGA
{
    VGARegistersFree(savedVGARegisters);
    VGARegistersFree(vgaRegisters);

    IOFree(savedRegisters, sizeof(NeoMagicRegisters));
    IOFree(registers, sizeof(NeoMagicRegisters));
    IOFree(savedVGARegisters, sizeof(VGARegisters));
    IOFree(vgaRegisters, sizeof(VGARegisters));

    IOFree(VGA, sizeof(VGADescription));
}

- (void)lock
{
    VGA->writeGraphics(VGA, 0x09, 0x00);
    VGALock(VGA);
}

- (void)unlock
{
    VGAUnlock(VGA);
    VGA->writeGraphics(VGA, 0x09, 0x26);
}

- (void)saveMode
{
    int i;
    
    VGA->writeGraphics(VGA, 0x09, 0x26);
    VGA->writeGraphics(VGA, 0x15, 0x00);  /* Get back to bank zero. */
    
    VGASave(VGA, savedVGARegisters);
    
    /*
     * Extended Graphics Registers
     */
    savedRegisters->generalLock = VGA->readGraphics(VGA, 0x0A);
    savedRegisters->externalCRTDisplayAddress = VGA->readGraphics(VGA, 0x0A);
    if (chipset != NM2070)
        savedRegisters->externalCRTOffset = VGA->readGraphics(VGA, 0x0F);
    savedRegisters->systemInterfaceControl1 = VGA->readGraphics(VGA, 0x10);
    savedRegisters->systemInterfaceControl2 = VGA->readGraphics(VGA, 0x11);
    savedRegisters->singleAddressPage = VGA->readGraphics(VGA, 0x15);
    savedRegisters->dualAddressPage = VGA->readGraphics(VGA, 0x16);
    savedRegisters->panelDisplayControl1 = VGA->readGraphics(VGA, 0x20);
    savedRegisters->panelDisplayControl2 = VGA->readGraphics(VGA, 0x25);
    savedRegisters->panelDisplayControl3 = VGA->readGraphics(VGA, 0x30);
    savedRegisters->panelVerticalCenter1 = VGA->readGraphics(VGA, 0x28);
    savedRegisters->panelVerticalCenter2 = VGA->readGraphics(VGA, 0x29);
    savedRegisters->panelVerticalCenter3 = VGA->readGraphics(VGA, 0x2A);
    if (chipset != NM2070) {
        savedRegisters->panelVerticalCenter4 = VGA->readGraphics(VGA, 0x32);
        savedRegisters->panelHorizontalCenter1 = VGA->readGraphics(VGA, 0x33);
        savedRegisters->panelHorizontalCenter2 = VGA->readGraphics(VGA, 0x34);
        savedRegisters->panelHorizontalCenter3 = VGA->readGraphics(VGA, 0x35);
    }
    if (chipset == NM2160)
        savedRegisters->panelHorizontalCenter4 = VGA->readGraphics(VGA, 0x36);
    else if (chipset == NM2200 || chipset == NM2230 ||
             chipset == NM2360 || chipset == NM2380) {
        savedRegisters->panelHorizontalCenter4 = VGA->readGraphics(VGA, 0x36);
        savedRegisters->panelVerticalCenter5 = VGA->readGraphics(VGA, 0x37);
        savedRegisters->panelHorizontalCenter5 = VGA->readGraphics(VGA, 0x38);
    }
    savedRegisters->extendedColorModeSelect = VGA->readGraphics(VGA, 0x90);
    savedRegisters->VCLK3NumeratorLow = VGA->readGraphics(VGA, 0x9B);
    if (chipset == NM2200 || chipset == NM2230 ||
        chipset == NM2360 || chipset == NM2380)
	savedRegisters->VCLK3NumeratorHigh = VGA->readGraphics(VGA, 0x8F);
    savedRegisters->VCLK3Denominator = VGA->readGraphics(VGA, 0x9F);
    savedRegisters->programVCLK = 0x01;
    
    /*
     * CRTC Controller
     */
    savedRegisters->CRTC[0x23] = VGA->readCRTC(VGA, 0x23);
    savedRegisters->CRTC[0x25] = VGA->readCRTC(VGA, 0x25);
    savedRegisters->CRTC[0x2F] = VGA->readCRTC(VGA, 0x2F);
    for (i = 0x40; i <= 0x59; i++)
        savedRegisters->CRTC[i] = VGA->readCRTC(VGA, i);
    for (i = 0x60; i <= 0x69; i++)
        savedRegisters->CRTC[i] = VGA->readCRTC(VGA, i);
    for (i = 0x70; i <= NEO_EXT_CRTC_MAX; i++)
        savedRegisters->CRTC[i] = VGA->readCRTC(VGA, i);
    
    /*
     * Graphics Display Controller
     */
    for (i = 0x0A; i <= NEO_EXT_GRAPHICS_MAX; i++)
        savedRegisters->graphics[i] = VGA->readGraphics(VGA, i);

    savedRegisters->useExtendedRegisters = 0x01;

}

- (void)programShadowRegisters:(NeoMagicRegisters *)reg
{
    unsigned char programShadowRegs;
    unsigned char programLCD;
    
    int i;
    
    programLCD = ((VGA->readGraphics(VGA, 0x20) & 0x03) == 0x01) ? 0x01 : 0x00;
    
    switch (chipset) {
        case NM2070:
	    programShadowRegs = 0x01;
	    
	    if (reg->panelDisplayControl2 & 0x84) {
	        /* Don't program by default if in stretch mode. */
		programShadowRegs = 0x00;
	    }
	    break;
	case NM2090:
	case NM2093:
	case NM2097:
	case NM2160:
	case NM2200:
	case NM2230:
	case NM2360:
	case NM2380:
	default:
	    /* Don't program the shadow registers by default. */
	    programShadowRegs = 0x00;
	    if (programLCD)
	        programShadowRegs = 0x01;
	    break;
    }
    
    if (!programShadowRegs) {
	if (savedRegisters->useExtendedRegisters) {
	    for (i = 0x40; i <= 0x59; i++)
	        VGA->writeCRTC(VGA, i, savedRegisters->CRTC[i]);
	    for (i = 0x60; i <= 0x64; i++)
	        VGA->writeCRTC(VGA, i, savedRegisters->CRTC[i]);
	}
    } else {
        switch (panelWidth) {
	    case 640:
	        VGA->writeCRTC(VGA, 0x40, 0x5F);
	        VGA->writeCRTC(VGA, 0x41, 0x50);
	        VGA->writeCRTC(VGA, 0x42, 0x02);
	        VGA->writeCRTC(VGA, 0x43, 0x55);
	        VGA->writeCRTC(VGA, 0x44, 0x81);
	        VGA->writeCRTC(VGA, 0x45, 0x0B);
	        VGA->writeCRTC(VGA, 0x46, 0x2E);
	        VGA->writeCRTC(VGA, 0x47, 0xEA);
	        VGA->writeCRTC(VGA, 0x48, 0x0C);
	        VGA->writeCRTC(VGA, 0x49, 0xE7);
	        VGA->writeCRTC(VGA, 0x4A, 0x04);
	        VGA->writeCRTC(VGA, 0x4B, 0x2D);
	        VGA->writeCRTC(VGA, 0x4C, 0x28);
	        VGA->writeCRTC(VGA, 0x4D, 0x90);
	        VGA->writeCRTC(VGA, 0x4E, 0x2B);
	        VGA->writeCRTC(VGA, 0x4F, 0xA0);
	        break;
	    case 800:
	        switch (panelHeight) {
		    case 600:
	                VGA->writeCRTC(VGA, 0x40, 0x7F);
	                VGA->writeCRTC(VGA, 0x41, 0x63);
	                VGA->writeCRTC(VGA, 0x42, 0x02);
	                VGA->writeCRTC(VGA, 0x43, 0x6C);
	                VGA->writeCRTC(VGA, 0x44, 0x1C);
	                VGA->writeCRTC(VGA, 0x45, 0x72);
	                VGA->writeCRTC(VGA, 0x46, 0xE0);
	                VGA->writeCRTC(VGA, 0x47, 0x58);
	                VGA->writeCRTC(VGA, 0x48, 0x0C);
	                VGA->writeCRTC(VGA, 0x49, 0x57);
	                VGA->writeCRTC(VGA, 0x4A, 0x73);
	                VGA->writeCRTC(VGA, 0x4B, 0x3D);
	                VGA->writeCRTC(VGA, 0x4C, 0x31);
	                VGA->writeCRTC(VGA, 0x4D, 0x01);
	                VGA->writeCRTC(VGA, 0x4E, 0x36);
	                VGA->writeCRTC(VGA, 0x4F, 0x1E);
		        if (chipset != NM2070) {
	                    VGA->writeCRTC(VGA, 0x50, 0x6B);
	                    VGA->writeCRTC(VGA, 0x51, 0x4F);
	                    VGA->writeCRTC(VGA, 0x52, 0x0E);
	                    VGA->writeCRTC(VGA, 0x53, 0x58);
	                    VGA->writeCRTC(VGA, 0x54, 0x88);
	                    VGA->writeCRTC(VGA, 0x55, 0x33);
	                    VGA->writeCRTC(VGA, 0x56, 0x27);
	                    VGA->writeCRTC(VGA, 0x57, 0x16);
	                    VGA->writeCRTC(VGA, 0x58, 0x2C);
	                    VGA->writeCRTC(VGA, 0x59, 0x94);
		        }
			break;
		    case 480:
	                VGA->writeCRTC(VGA, 0x40, 0x7F);
	                VGA->writeCRTC(VGA, 0x41, 0x63);
	                VGA->writeCRTC(VGA, 0x42, 0x02);
	                VGA->writeCRTC(VGA, 0x43, 0x6B);
	                VGA->writeCRTC(VGA, 0x44, 0x1B);
	                VGA->writeCRTC(VGA, 0x45, 0x72);
	                VGA->writeCRTC(VGA, 0x46, 0xE0);
	                VGA->writeCRTC(VGA, 0x47, 0x1C);
	                VGA->writeCRTC(VGA, 0x48, 0x00);
	                VGA->writeCRTC(VGA, 0x49, 0x57);
	                VGA->writeCRTC(VGA, 0x4A, 0x73);
	                VGA->writeCRTC(VGA, 0x4B, 0x3E);
	                VGA->writeCRTC(VGA, 0x4C, 0x31);
	                VGA->writeCRTC(VGA, 0x4D, 0x01);
	                VGA->writeCRTC(VGA, 0x4E, 0x36);
	                VGA->writeCRTC(VGA, 0x4F, 0x1E);
	                VGA->writeCRTC(VGA, 0x50, 0x6B);
	                VGA->writeCRTC(VGA, 0x51, 0x4F);
	                VGA->writeCRTC(VGA, 0x52, 0x0E);
	                VGA->writeCRTC(VGA, 0x53, 0x57);
	                VGA->writeCRTC(VGA, 0x54, 0x87);
	                VGA->writeCRTC(VGA, 0x55, 0x33);
	                VGA->writeCRTC(VGA, 0x56, 0x27);
	                VGA->writeCRTC(VGA, 0x57, 0x16);
	                VGA->writeCRTC(VGA, 0x58, 0x2C);
	                VGA->writeCRTC(VGA, 0x59, 0x94);
		        break;
		}
	        break;
	    case 1024:
	        switch (panelHeight) {
		    case 768:
	                VGA->writeCRTC(VGA, 0x40, 0xA3);
	                VGA->writeCRTC(VGA, 0x41, 0x7F);
	                VGA->writeCRTC(VGA, 0x42, 0x06);
	                VGA->writeCRTC(VGA, 0x43, 0x85);
	                VGA->writeCRTC(VGA, 0x44, 0x96);
	                VGA->writeCRTC(VGA, 0x45, 0x24);
	                VGA->writeCRTC(VGA, 0x46, 0xE5);
	                VGA->writeCRTC(VGA, 0x47, 0x02);
	                VGA->writeCRTC(VGA, 0x48, 0x08);
	                VGA->writeCRTC(VGA, 0x49, 0xFF);
	                VGA->writeCRTC(VGA, 0x4A, 0x25);
	                VGA->writeCRTC(VGA, 0x4B, 0x4F);
	                VGA->writeCRTC(VGA, 0x4C, 0x40);
	                VGA->writeCRTC(VGA, 0x4D, 0x00);
	                VGA->writeCRTC(VGA, 0x4E, 0x44);
	                VGA->writeCRTC(VGA, 0x4F, 0x0C);
	                VGA->writeCRTC(VGA, 0x50, 0x7A);
	                VGA->writeCRTC(VGA, 0x51, 0x56);
	                VGA->writeCRTC(VGA, 0x52, 0x00);
	                VGA->writeCRTC(VGA, 0x53, 0x5D);
	                VGA->writeCRTC(VGA, 0x54, 0x0E);
	                VGA->writeCRTC(VGA, 0x55, 0x3B);
	                VGA->writeCRTC(VGA, 0x56, 0x2B);
	                VGA->writeCRTC(VGA, 0x57, 0x00);
	                VGA->writeCRTC(VGA, 0x58, 0x2F);
	                VGA->writeCRTC(VGA, 0x59, 0x18);
	                VGA->writeCRTC(VGA, 0x60, 0x88);
	                VGA->writeCRTC(VGA, 0x61, 0x63);
	                VGA->writeCRTC(VGA, 0x62, 0x0B);
	                VGA->writeCRTC(VGA, 0x63, 0x69);
	                VGA->writeCRTC(VGA, 0x64, 0x1A);
			break;
		    case 480:
	                VGA->writeCRTC(VGA, 0x40, 0xA3);
	                VGA->writeCRTC(VGA, 0x41, 0x7F);
	                VGA->writeCRTC(VGA, 0x42, 0x1B);
	                VGA->writeCRTC(VGA, 0x43, 0x89);
	                VGA->writeCRTC(VGA, 0x44, 0x16);
	                VGA->writeCRTC(VGA, 0x45, 0x0B);
	                VGA->writeCRTC(VGA, 0x46, 0x2C);
	                VGA->writeCRTC(VGA, 0x47, 0xE8);
	                VGA->writeCRTC(VGA, 0x48, 0x0C);
	                VGA->writeCRTC(VGA, 0x49, 0xE7);
	                VGA->writeCRTC(VGA, 0x4A, 0x09);
	                VGA->writeCRTC(VGA, 0x4B, 0x4F);
	                VGA->writeCRTC(VGA, 0x4C, 0x40);
	                VGA->writeCRTC(VGA, 0x4D, 0x00);
	                VGA->writeCRTC(VGA, 0x4E, 0x44);
	                VGA->writeCRTC(VGA, 0x4F, 0x0C);
	                VGA->writeCRTC(VGA, 0x50, 0x7A);
	                VGA->writeCRTC(VGA, 0x51, 0x56);
	                VGA->writeCRTC(VGA, 0x52, 0x00);
	                VGA->writeCRTC(VGA, 0x53, 0x5D);
	                VGA->writeCRTC(VGA, 0x54, 0x0E);
	                VGA->writeCRTC(VGA, 0x55, 0x3B);
	                VGA->writeCRTC(VGA, 0x56, 0x2A);
	                VGA->writeCRTC(VGA, 0x57, 0x00);
	                VGA->writeCRTC(VGA, 0x58, 0x2F);
	                VGA->writeCRTC(VGA, 0x59, 0x18);
	                VGA->writeCRTC(VGA, 0x60, 0x88);
	                VGA->writeCRTC(VGA, 0x61, 0x63);
	                VGA->writeCRTC(VGA, 0x62, 0x0B);
	                VGA->writeCRTC(VGA, 0x63, 0x69);
	                VGA->writeCRTC(VGA, 0x64, 0x1A);
			break;
		}
	        break;
	}
    }
}

- (void)restoreModeWithRegisters:(NeoMagicRegisters *)reg
                    VGARegisters:(VGARegisters *)vgareg
{
    unsigned char tmp;
    unsigned char clockHi = 0x00;
    
    int i;
    
    [self programShadowRegisters:reg];

    VGA->writeGraphics(VGA, 0x15, 0x00);
    VGA->writeGraphics(VGA, 0x0A, reg->generalLock);
    
    /*
     * The color mode needs to be set before calling VGARestore to
     * ensure the DAC is initilized properly.
     */
    tmp = VGA->readGraphics(VGA, 0x90);
    
    switch (chipset) {
        case NM2070:
	    tmp &= 0xF0;  /* Save bits 7:4. */
	    tmp |= (reg->extendedColorModeSelect & ~0xF0);
	    break;
	case NM2090:
	case NM2093:
	case NM2097:
	case NM2160:
	case NM2200:
	case NM2230:
	case NM2360:
	case NM2380:
	    tmp &= 0x70;  /* Save bits 6:4. */
	    tmp |= (reg->extendedColorModeSelect & ~0x70);
	    break;
    }
    VGA->writeGraphics(VGA, 0x90, tmp);
    
    /* Prevent a lock. */
    IODelay(200000);
    
    /*
     * Disable horizontal and vertical graphics and text expansion so
     * VGARestore works properly.
     */
    tmp = VGA->readGraphics(VGA, 0x25);
    tmp &= 0x39;
    VGA->writeGraphics(VGA, 0x25, tmp);
    
    /* Wait for the above ops to take effect. */
    IODelay(200000);
    
    /* Restore the standard VGA registers. */
    VGARestore(VGA, vgareg);
    
    VGA->writeGraphics(VGA, 0x0E, reg->externalCRTDisplayAddress);
    VGA->writeGraphics(VGA, 0x0F, reg->externalCRTOffset);
    tmp = VGA->readGraphics(VGA, 0x10);
    tmp &= 0x0F;  /* Save bits 3:0. */
    tmp |= (reg->systemInterfaceControl1 & ~0x0F);
    VGA->writeGraphics(VGA, 0x10, tmp);
    
    VGA->writeGraphics(VGA, 0x11, reg->systemInterfaceControl2);
    VGA->writeGraphics(VGA, 0x15, reg->singleAddressPage);
    VGA->writeGraphics(VGA, 0x16, reg->dualAddressPage);
    
    tmp = VGA->readGraphics(VGA, 0x20);
    switch (chipset) {
        case NM2070:
	    tmp &= 0xFC;  /* Save bits 7:2. */
	    tmp |= (reg->panelDisplayControl1 & ~0xFC);
	    break;
	case NM2090:
	case NM2093:
	case NM2097:
	case NM2160:
	    tmp &= 0xDC;  /* Save bits 7:6, 4:2. */
	    tmp |= (reg->panelDisplayControl1 & ~0xDC);
	    break;
	case NM2200:
	case NM2230:
	case NM2360:
	case NM2380:
	    tmp &= 0x98;  /* Save bits 7, 4:3. */
	    tmp |= (reg->panelDisplayControl1 & ~0x98);
	    break;
    }
    VGA->writeGraphics(VGA, 0x20, tmp);
    
    tmp = VGA->readGraphics(VGA, 0x25);
    tmp &= 0x38;  /* Save bits 5:3. */
    tmp |= (reg->panelDisplayControl2 & ~0x38);
    VGA->writeGraphics(VGA, 0x25, tmp);
    
    if (chipset != NM2070) {
        tmp = VGA->readGraphics(VGA, 0x30);
        tmp &= 0xEF;  /* Save bits 7:5, 3:0. */
        tmp |= (reg->panelDisplayControl3 & ~0xEF);
        VGA->writeGraphics(VGA, 0x30, tmp);
    }
    
    VGA->writeGraphics(VGA, 0x28, reg->panelVerticalCenter1);
    VGA->writeGraphics(VGA, 0x29, reg->panelVerticalCenter2);
    VGA->writeGraphics(VGA, 0x2A, reg->panelVerticalCenter3);
    
    if (chipset != NM2070) {
        VGA->writeGraphics(VGA, 0x32, reg->panelVerticalCenter4);
        VGA->writeGraphics(VGA, 0x33, reg->panelHorizontalCenter1);
        VGA->writeGraphics(VGA, 0x34, reg->panelHorizontalCenter2);
        VGA->writeGraphics(VGA, 0x35, reg->panelHorizontalCenter3);
    }
    
    if (chipset == NM2160)
        VGA->writeGraphics(VGA, 0x36, reg->panelHorizontalCenter4);
    if (chipset == NM2200 || chipset == NM2230 ||
             chipset == NM2360 || chipset == NM2380) {
        VGA->writeGraphics(VGA, 0x36, reg->panelHorizontalCenter4);
        VGA->writeGraphics(VGA, 0x37, reg->panelVerticalCenter5);
        VGA->writeGraphics(VGA, 0x38, reg->panelHorizontalCenter5);
    }
    
    if (chipset == NM2200 || chipset == NM2230 ||
        chipset == NM2360 || chipset == NM2380)
	clockHi = 0x01;

    /* Program VCLK3 if needed. */
    if (reg->programVCLK &&
        ((VGA->readGraphics(VGA, 0x9B) != reg->VCLK3NumeratorLow) ||
         (VGA->readGraphics(VGA, 0x9F) != reg->VCLK3Denominator) ||
	 (clockHi && ((VGA->readGraphics(VGA, 0x8F) & ~0x0F) !=
	              (reg->VCLK3NumeratorHigh & ~0x0F))))) {
        VGA->writeGraphics(VGA, 0x9B, reg->VCLK3NumeratorLow);
	if (clockHi) {
	    tmp = VGA->readGraphics(VGA, 0x8F);
	    tmp &= 0x0F;  /* Save bits 3:0. */
	    tmp |= (reg->VCLK3NumeratorHigh & ~0x0F);
	    VGA->writeGraphics(VGA, 0x8F, tmp);
	}
	VGA->writeGraphics(VGA, 0x9F, reg->VCLK3Denominator);
    }
    if (reg->biosMode)
        VGA->writeCRTC(VGA, 0x23, reg->biosMode);

    if (reg->useExtendedRegisters) {
        VGA->writeCRTC(VGA, 0x23, reg->CRTC[0x23]);	
        VGA->writeCRTC(VGA, 0x25, reg->CRTC[0x25]);	
        VGA->writeCRTC(VGA, 0x2F, reg->CRTC[0x2F]);
        for (i = 0x40; i <= 0x59; i++)
            VGA->writeCRTC(VGA, i, reg->CRTC[i]);
        for (i = 0x60; i <= 0x69; i++)
            VGA->writeCRTC(VGA, i, reg->CRTC[i]);
        for (i = 0x70; i <= NEO_EXT_CRTC_MAX; i++)
            VGA->writeCRTC(VGA, i, reg->CRTC[i]);

        for (i = 0x0A; i <= 0x3F; i++)
            VGA->writeGraphics(VGA, i, reg->graphics[i]);
        for (i = 0x90; i <= NEO_EXT_GRAPHICS_MAX; i++)
            VGA->writeGraphics(VGA, i, reg->graphics[i]);
    }

    /* Program vertical extension register. */
    if (chipset == NM2200 || chipset == NM2230 ||
        chipset == NM2360 || chipset == NM2380)
	VGA->writeCRTC(VGA, 0x70, reg->verticalExtension);
}

- (void)setGammaTable
{
    unsigned int i, j, g;
    const IODisplayInfo *displayInfo;

    displayInfo = [self displayInfo];

    if (redTransferTable != 0) {
	for (i = 0; i < transferTableCount; i++)
	    SetGammaValue(VGA, vgaMode, i,
		          redTransferTable[i], greenTransferTable[i],
			  blueTransferTable[i], brightnessLevel);
    } else {
	switch (displayInfo->bitsPerPixel) {
	case IO_24BitsPerPixel:
	case IO_8BitsPerPixel:
	    for (g = 0; g < 256; g++) {
		SetGammaValue(VGA, vgaMode, g,
		              gamma8[g], gamma8[g], gamma8[g],
			      brightnessLevel);
	    }
	    break;
	case IO_15BitsPerPixel:
	    for (i = 0; i < 32; i++) {
		for (j = 0; j < 8; j++) {
		    SetGammaValue(VGA, vgaMode, i * 8 + j,
		                  gamma16[i/2], gamma16[i/2], gamma16[i/2],
				  brightnessLevel);
		}
	    }
	    break;
	default:
	    break;
	}
    }
}

#define MAX_N 127
#define MAX_D 31
#define MAX_F 1

- (void)calcVCLKFromFreq:(long)freq {
    
    int n, d, f;
    int n_best = 0, d_best = 0, f_best = 0;
    long f_best_diff = 0x7ffff;
    
    for (f = 0; f <= MAX_F; f++)
        for (d = 0; d <= MAX_D; d++)
            for (n = 0; n <= MAX_N; n++) {
                long f_out;
                long f_diff;
                
                f_out = ((14318 * (n + 1)) / (d + 1)) >> f;
                f_diff = abs(f_out - freq);
                if (f_diff <= f_best_diff) {
                    f_best_diff = f_diff;
                    n_best = n;
                    d_best = d;
                    f_best = f;
                }
                if (f_out > freq)
                    break;
            }
    
    if (chipset == NM2200 ||
        chipset == NM2230 ||
        chipset == NM2360 ||
        chipset == NM2380) {
        /* NOT_DONE:  We are trying the full range of the 2200 clock.
         We should be able to try n up to 2047 */
        registers->VCLK3NumeratorLow = n_best;
        registers->VCLK3NumeratorHigh = (f_best << 7);
    } else
        registers->VCLK3NumeratorLow = n_best | (f_best << 7);
    
    registers->VCLK3Denominator = d_best;
    //IOLog("VCLK: d_best: %d, n_best: %d, f_best: %d, f_best_diff: %f, f_target: %f, orig_freq: %ld\n", d_best, n_best, f_best, f_best_diff, f_target, freq);
}

@end
