/*
 * VGA.c
 * VGA Controller Manipulation
 *
 * Copyright (c) 2004 David Lazaro Saz.
 * All rights reserved.
 */

#import "VGA.h"

#import <driverkit/i386/ioPorts.h>
#import <driverkit/generalFuncs.h>


#ifndef NULL
#define NULL ((void *)0)
#endif


/*
 * Register access functions
 */
static unsigned char VGAReadCRTC(VGADescription *VGA, unsigned char index)
{
    outb(VGA->IOBase + VGA->PIOOffset + VGA_CRTC_INDEX_OFFSET, index);
    return inb(VGA->IOBase + VGA->PIOOffset + VGA_CRTC_DATA_OFFSET);
}

static void VGAWriteCRTC(VGADescription *VGA,
                         unsigned char index, unsigned char value)
{
    outb(VGA->IOBase + VGA->PIOOffset + VGA_CRTC_INDEX_OFFSET, index);
    outb(VGA->IOBase + VGA->PIOOffset + VGA_CRTC_DATA_OFFSET, value);
}

static unsigned char VGAReadGraphics(VGADescription *VGA, unsigned char index)
{
    outb(VGA->PIOOffset + VGA_GRAPH_INDEX, index);
    return inb(VGA->PIOOffset + VGA_GRAPH_DATA);
}

static void VGAWriteGraphics(VGADescription *VGA,
                             unsigned char index, unsigned char value)
{
    outb(VGA->PIOOffset + VGA_GRAPH_INDEX, index);
    outb(VGA->PIOOffset + VGA_GRAPH_DATA, value);
}

static unsigned char VGAReadSequencer(VGADescription *VGA, unsigned char index)
{
    outb(VGA->PIOOffset + VGA_SEQ_INDEX, index);
    return inb(VGA->PIOOffset + VGA_SEQ_DATA);
}

static void VGAWriteSequencer(VGADescription *VGA,
                              unsigned char index, unsigned char value)
{
    outb(VGA->PIOOffset + VGA_SEQ_INDEX, index);
    outb(VGA->PIOOffset + VGA_SEQ_DATA, value);
}

static unsigned char VGAReadAttribute(VGADescription *VGA, unsigned char index)
{
    unsigned char tmp;
    
    if (VGA->paletteEnabled)
        index &= ~0x20;
    else
        index |= 0x20;

    tmp = inb(VGA->IOBase + VGA->PIOOffset + VGA_IN_STAT_1_OFFSET);
    outb(VGA->PIOOffset + VGA_ATTR_INDEX, index);
    return inb(VGA->PIOOffset + VGA_ATTR_DATA_R);
}

static void VGAWriteAttribute(VGADescription *VGA,
                              unsigned char index, unsigned char value)
{
    unsigned char tmp;
    
    if (VGA->paletteEnabled)
        index &= ~0x20;
    else
        index |= 0x20;

    tmp = inb(VGA->IOBase + VGA->PIOOffset + VGA_IN_STAT_1_OFFSET);
    outb(VGA->PIOOffset + VGA_ATTR_INDEX, index);
    outb(VGA->PIOOffset + VGA_ATTR_DATA_W, value);
}

static unsigned char VGAReadMiscOut(VGADescription *VGA)
{
    return inb(VGA->PIOOffset + VGA_MISC_OUT_R);
}

static void VGAWriteMiscOut(VGADescription *VGA, unsigned char value)
{
    outb(VGA->PIOOffset + VGA_MISC_OUT_W, value);
}

static void VGAEnablePalette(VGADescription *VGA)
{
    unsigned char tmp;
    
    tmp = inb(VGA->IOBase + VGA->PIOOffset + VGA_IN_STAT_1_OFFSET);
    outb(VGA->PIOOffset + VGA_ATTR_INDEX, 0x00);
    VGA->paletteEnabled = 0x01;
}

static void VGADisablePalette(VGADescription *VGA)
{
    unsigned char tmp;
    
    tmp = inb(VGA->IOBase + VGA->PIOOffset + VGA_IN_STAT_1_OFFSET);
    outb(VGA->PIOOffset + VGA_ATTR_INDEX, 0x20);
    VGA->paletteEnabled = 0x00;
}

static void VGAWriteDACWriteAddress(VGADescription *VGA, unsigned char value)
{
    outb(VGA->PIOOffset + VGA_DAC_WRITE_ADDR, value);
}

static void VGAWriteDACData(VGADescription *VGA, unsigned char value)
{
    outb(VGA->PIOOffset + VGA_DAC_DATA, value);
}


/*
 * Configuration functions
 */
unsigned char VGAGetIOBase(VGADescription *VGA)
{
    return (VGA->readMiscOut(VGA) & 0x01) ? VGA_IOBASE_COLOR : VGA_IOBASE_MONO;
}


void VGASequencerReset(VGADescription *VGA, unsigned char start)
{
    if (start)
        VGA->writeSequencer(VGA, 0x00, 0x01);
    else
        VGA->writeSequencer(VGA, 0x00, 0x03);
}

void VGAProtect(VGADescription *VGA)
{
    unsigned char tmp = VGA->readSequencer(VGA, 0x01);
    
    VGASequencerReset(VGA, 0x01);                /* Start synchronous reset. */
    VGA->writeSequencer(VGA, 0x01, tmp | 0x20);  /* Disable the display. */
    
    VGA->enablePalette(VGA);
}

void VGAUnprotect(VGADescription *VGA)
{
    unsigned char tmp = VGA->readSequencer(VGA, 0x01);
    
    VGA->writeSequencer(VGA, 0x01, tmp & ~0x20); /* Reenable the display. */
    VGASequencerReset(VGA, 0x00);                /* Clear synchronous reset. */
    
    VGA->disablePalette(VGA);
}

void VGALock(VGADescription *VGA)
{
    /* Protect CRTC[0-7]. */
    VGA->writeCRTC(VGA, 0x11, VGA->readCRTC(VGA, 0x11) | 0x80);
}

void VGAUnlock(VGADescription *VGA)
{
    /* Unprotect CRTC[0-7]. */
    VGA->writeCRTC(VGA, 0x11, VGA->readCRTC(VGA, 0x11) | ~0x80);
}


void VGARegistersAlloc(VGARegisters *registers)
{
    unsigned char *buf;
    
    buf = IOMalloc(registers->numCRTC +
                  registers->numSequencer +
		  registers->numGraphics +
		  registers->numAttribute);
    registers->CRTC = buf;
    registers->sequencer = registers->CRTC + registers->numCRTC;
    registers->graphics = registers->sequencer + registers->numSequencer;
    registers->attribute = registers->graphics + registers->numGraphics;
}

void VGARegistersFree(VGARegisters *registers)
{
    unsigned char *buf;
    
    buf = registers->CRTC;
    IOFree(buf, registers->numCRTC + registers->numSequencer +
                registers->numGraphics + registers->numAttribute);
}

/*
 * VGAGetDescription
 * Get a VGA device caracterization.
 */
void VGAGetDescription(VGADescription *VGA, VGARegisters *registers)
{
    int i;
    
    VGA->readCRTC             = VGAReadCRTC;
    VGA->writeCRTC            = VGAWriteCRTC;
    VGA->readGraphics         = VGAReadGraphics;
    VGA->writeGraphics        = VGAWriteGraphics;
    VGA->readSequencer        = VGAReadSequencer;
    VGA->writeSequencer       = VGAWriteSequencer;
    VGA->readAttribute        = VGAReadAttribute;
    VGA->writeAttribute       = VGAWriteAttribute;
    VGA->readMiscOut          = VGAReadMiscOut;
    VGA->writeMiscOut         = VGAWriteMiscOut;
    VGA->enablePalette        = VGAEnablePalette;
    VGA->disablePalette       = VGADisablePalette;
    VGA->writeDACWriteAddress = VGAWriteDACWriteAddress;
    VGA->writeDACData         = VGAWriteDACData;
    
    VGA->PIOOffset = 0;
    VGA->IOBase = VGAGetIOBase(VGA);
    
    VGA->paletteEnabled = 0x00;
    
    /* Set all colors to black. */
    for (i = 0; i < 768; i++)
        registers->DAC[i] = 0x00;
}

/*
 * VGASave
 * Save the current VGA state.
 */
void VGASave(VGADescription *VGA, VGARegisters *savedRegisters)
{
    int i;
    
    if (savedRegisters == NULL)
        return;

    savedRegisters->miscOut = VGA->readMiscOut(VGA);
    if (savedRegisters->miscOut & 0x01)
        VGA->IOBase = VGA_IOBASE_COLOR;
    else
        VGA->IOBase = VGA_IOBASE_MONO;
    
    for (i = 0; i < savedRegisters->numCRTC; i++)
        savedRegisters->CRTC[i] = VGA->readCRTC(VGA, i);

    VGA->enablePalette(VGA);
    for (i = 0; i < savedRegisters->numAttribute; i++)
        savedRegisters->attribute[i] = VGA->readAttribute(VGA, i);
    VGA->disablePalette(VGA);

    for (i = 0; i < savedRegisters->numGraphics; i++)
        savedRegisters->graphics[i] = VGA->readGraphics(VGA, i);

    for (i = 0; i < savedRegisters->numSequencer; i++)
        savedRegisters->graphics[i] = VGA->readSequencer(VGA, i);
}


/*
 * VGARestore
 * Restore a saved VGA state.
 */
void VGARestore(VGADescription *VGA, VGARegisters *restoredRegisters)
{
    int i;
    
    if (restoredRegisters->miscOut & 0x01)
        VGA->IOBase = VGA_IOBASE_COLOR;
    else
        VGA->IOBase = VGA_IOBASE_MONO;

    VGA->writeMiscOut(VGA, restoredRegisters->miscOut);
    
    for (i = 0; i < restoredRegisters->numSequencer; i++)
        VGA->writeSequencer(VGA, i, restoredRegisters->sequencer[i]);

    /* Ensure CRTC registers 0-7 are unlocked by clearing bit 7 or CRTC[17]. */
    VGA->writeCRTC(VGA, 17, restoredRegisters->CRTC[17] & ~0x80);
    
    for (i = 0; i < restoredRegisters->numCRTC; i++)
        VGA->writeCRTC(VGA, i, restoredRegisters->CRTC[i]);

    for (i = 0; i < restoredRegisters->numGraphics; i++)
        VGA->writeGraphics(VGA, i, restoredRegisters->graphics[i]);

    VGA->enablePalette(VGA);
    for (i = 0; i < restoredRegisters->numAttribute; i++)
        VGA->writeAttribute(VGA, i, restoredRegisters->attribute[i]);
    VGA->disablePalette(VGA);
}

void VGAEnterMode(VGADescription *VGA, VGARegisters *registers,
                  const VGAMode *mode)
{
    int i;
    
    if ((mode->flags & (VGAModePHSync | VGAModeNHSync)) &&
        (mode->flags & (VGAModePVSync | VGAModeNVSync))) {
        registers->miscOut = 0x23;
	if (mode->flags & VGAModeNHSync)
	    registers->miscOut |= 0x40;
	if (mode->flags & VGAModeNVSync)
	    registers->miscOut |= 0x80;
    } else {
        int vDisplay = mode->vDisplay;
	
	if (mode->flags & VGAModeDoubleScan)
	    vDisplay *= 2;
	if (mode->vScan > 1)
	    vDisplay *= mode->vScan;
	    
	if (vDisplay < 400)
	    registers->miscOut = 0xA3;  /* +HSync -VSync */
	else if (vDisplay < 480)
	    registers->miscOut = 0x63;  /* -HSync +VSync */
	else if (vDisplay < 768)
	    registers->miscOut = 0xE3;  /* -HSync -VSync */
	else
	    registers->miscOut = 0x23;  /* +HSymc +VSync */
    }
    
    registers->miscOut |= (mode->clockIndex & 0x03) << 2;
    
    /*
     * Time Sequencer
     */
    if (mode->depth == 4)
        registers->sequencer[0] = 0x02;
    else
        registers->sequencer[0] = 0x00;

    if (mode->flags & VGAModeClockDiv2)
        registers->sequencer[1] = 0x09;
    else
        registers->sequencer[1] = 0x01;

    if (mode->depth == 1)
        registers->sequencer[2] = 1 << BIT_PLANE;
    else
        registers->sequencer[2] = 0x0F;

    registers->sequencer[3] = 0x00;  /* Font select. */
    
    if (mode->depth < 8)
        registers->sequencer[4] = 0x06;  /* Misc. */
    else
        registers->sequencer[4] = 0x0E;  /* Misc. */

    /*
     * CRTC Controller
     */
    registers->CRTC[0]  = (mode->hTotal >> 3) - 5;
    registers->CRTC[1]  = (mode->hDisplay >> 3) - 1;
    registers->CRTC[2]  = (mode->hBlankStart >> 3) - 1;
    registers->CRTC[3]  = (((mode->hBlankEnd >> 3) - 1) & 0x1F) | 0x80;
    
    i = (((mode->hSkew << 2) + 0x10) & ~0x1F);
    if (i < 0x80)
        registers->CRTC[3] |= i;

    registers->CRTC[4]  = (mode->hSyncStart >> 3);
    registers->CRTC[5]  = ((((mode->hBlankEnd >> 3) - 1) & 0x20) << 2) |
                          (((mode->hSyncEnd >> 3)) & 0xFF);
    registers->CRTC[6]  = (mode->vTotal - 2) & 0xFF;
    registers->CRTC[7]  = (((mode->vTotal - 2) & 0x100) >> 8) |
                          (((mode->vDisplay - 1) & 0x100) >> 7) |
			  ((mode->vSyncStart & 0x100) >> 6) |
			  (((mode->vBlankStart - 1) & 0x100) >> 5) |
			  0x10 |
			  (((mode->vTotal - 2) & 0x200) >> 4) |
			  (((mode->vDisplay - 1) & 0x200) >> 3) |
		 	  ((mode->vSyncStart & 0x200) >> 2);
    registers->CRTC[8]  = 0x00;
    registers->CRTC[9]  = (((mode->vBlankStart - 1) & 0x200) >> 4) | 0x40;

    if (mode->flags & VGAModeDoubleScan)
        registers->CRTC[9] |= 0x80;

    if (mode->vScan >= 32)
        registers->CRTC[9] |= 0x1F;
    else if (mode->vScan > 1)
        registers->CRTC[9] |= mode->vScan - 1;

    registers->CRTC[10] = 0x00;
    registers->CRTC[11] = 0x00;
    registers->CRTC[12] = 0x00;
    registers->CRTC[13] = 0x00;
    registers->CRTC[14] = 0x00,
    registers->CRTC[15] = 0x00;
    registers->CRTC[16] = mode->vSyncStart & 0xFF;
    registers->CRTC[17] = (mode->vSyncEnd & 0x0F) | 0x20;
    registers->CRTC[18] = (mode->vDisplay - 1) & 0xFF;
    registers->CRTC[19] = mode->vDisplay >> 4;  /* Just a guess. */
    registers->CRTC[20] = 0x00;
    registers->CRTC[21] = (mode->vBlankStart - 1) & 0xFF;
    registers->CRTC[22] = (mode->vBlankEnd - 1) & 0xFF;
    
    if (mode->depth < 8)
        registers->CRTC[23] = 0xE3;
    else
        registers->CRTC[23] = 0xC3;

    registers->CRTC[24] = 0xFF;
    
    /*
     * Graphics Display Controller
     */
    registers->graphics[0] = 0x00;
    registers->graphics[1] = 0x00;
    registers->graphics[2] = 0x00;
    registers->graphics[3] = 0x00;
    
    if (mode->depth == 1) {
        registers->graphics[4] = BIT_PLANE;
	registers->graphics[5] = 0x00;
    } else {
        registers->graphics[4] = 0x00;
	if (mode->depth == 4)
	    registers->graphics[5] = 0x02;
	else
	    registers->graphics[5] = 0x40;
    }
    
    registers->graphics[6] = 0x05;  /* Only map 64KB of VGA memory. */
    registers->graphics[7] = 0x0F;
    registers->graphics[8] = 0xFF;
    
    /*
     * Video Attribute Controller
     */
    if (mode->depth == 1) {
        /* Initialize the mono map according to which bit-plane gets used. */
    } else {
        registers->attribute[0]  = 0x00;  /* Standard colormap translation. */
	registers->attribute[1]  = 0x01;
	registers->attribute[2]  = 0x02;
	registers->attribute[3]  = 0x03;
	registers->attribute[4]  = 0x04;
	registers->attribute[5]  = 0x05;
	registers->attribute[6]  = 0x06;
	registers->attribute[7]  = 0x07;
	registers->attribute[8]  = 0x08;
	registers->attribute[9]  = 0x09;
	registers->attribute[10] = 0x0A;
	registers->attribute[11] = 0x0B;
	registers->attribute[12] = 0x0C;
	registers->attribute[13] = 0x0D;
	registers->attribute[14] = 0x0E;
	registers->attribute[15] = 0x0F;
	
	if (mode->depth == 4)
	    registers->attribute[16] = 0x81;
	else
	    registers->attribute[16] = 0x41;
    }
    registers->attribute[17] = 0xFF;
    registers->attribute[18] = 0x0F;
    registers->attribute[19] = 0x00;
    registers->attribute[20] = 0x00;
}


/*
 * Register Logging
 */
void VGARegistersLog(const VGARegisters *reg, const char *module)
{
    int i;

    IOLog("%s: VGA Registers:\n", module);
    
    IOLog("%s:     Misc. Out = 0x%02x\n", module, reg->miscOut);
    
    IOLog("%s:     CRTC = {0x%02x", module, reg->CRTC[0]);
    for (i = 1; i < reg->numCRTC; i++) {
        if (!(i % 8))
	    IOLog("\n%s:             ", module);
        IOLog(", 0x%02x", reg->CRTC[i]);
    }
    IOLog("}\n");

    IOLog("%s:     SEQ = {0x%02x", module, reg->sequencer[0]);
    for (i = 1; i < reg->numSequencer; i++) {
        if (!(i % 8))
	    IOLog("\n%s:            ", module);
        IOLog(", 0x%02x", reg->sequencer[i]);
    }
    IOLog("}\n");
    
    IOLog("%s:     GFX = {0x%02x", module, reg->graphics[0]);
    for (i = 1; i < reg->numGraphics; i++) {
        if (!(i % 8))
	    IOLog("\n%s:            ", module);
        IOLog(", 0x%02x", reg->graphics[i]);
    }
    IOLog("}\n");

    IOLog("%s:     ATTR = {0x%02x", module, reg->attribute[0]);
    for (i = 1; i < reg->numAttribute; i++) {
        if (!(i % 8))
	    IOLog("\n%s:             ", module);
        IOLog(", 0x%02x", reg->attribute[i]);
    }
    IOLog("}\n");

    /* for (i = 0; i < 768; i++)
        IOLog("%s:     DAC[%d] = %02x\n", module, i, reg->DAC[i]); */
}
