/*
 * VGA.h
 * VGA Controller Manipulation
 *
 * Copyright (c) 2004 David Lazaro Saz.
 * All rights reserved.
 */

#ifndef VGA_H__
#define VGA_H__

#import <driverkit/i386/driverTypes.h>

/*
 * Standard VGA Registers
 */
#define VGA_ATTR_INDEX        0x3C0
#define VGA_ATTR_DATA_W       0x3C0
#define VGA_ATTR_DATA_R       0x3C1
#define VGA_IN_STAT_0         0x3C2  /* Read */
#define VGA_MISC_OUT_W        0x3C2  /* Write */
#define VGA_ENABLE            0x3C3
#define VGA_SEQ_INDEX         0x3C4
#define VGA_SEQ_DATA          0x3C5
#define VGA_DAC_MASK          0x3C6
#define VGA_DAC_READ_ADDR     0x3C7
#define VGA_DAC_WRITE_ADDR    0x3C8
#define VGA_DAC_DATA          0x3C9
#define VGA_FEATURE_R         0x3CA  /* Read */
#define VGA_MISC_OUT_R        0x3CC  /* Read */
#define VGA_GRAPH_INDEX       0x3CE
#define VGA_GRAPH_DATA        0x3CF

#define VGA_IOBASE_MONO       0x3B0
#define VGA_IOBASE_COLOR      0x3D0

#define VGA_CRTC_INDEX_OFFSET 0x04
#define VGA_CRTC_DATA_OFFSET  0x05
#define VGA_IN_STAT_1_OFFSET  0x0A  /* Read */
#define VGA_FEATURE_W_OFFSET  0x0A  /* Write */

#define OVERSCAN 0x11  /* Index of overscan register. */

#define BIT_PLANE 3  /* Which plane we write in mono mode. */

/*
 * Default number of VGA Registers stored internally
 */
#define VGA_NUM_CRTC 25
#define VGA_NUM_SEQ  5
#define VGA_NUM_GFX  9
#define VGA_NUM_ATTR 21


/*
 * VGA Registers Structure
 */
typedef struct {
    unsigned char  miscOut;
    unsigned char *CRTC;          /* CRT controller. */
    unsigned char *sequencer;     /* Video sequencer. */
    unsigned char *graphics;      /* Video graphics. */
    unsigned char *attribute;     /* Video attribute */
    unsigned char  DAC[768];      /* Internal color lookup table. */
    unsigned char  numCRTC;       /* Number of CRTC registers */
    unsigned char  numSequencer;  /* Number of sequencer registers. */
    unsigned char  numGraphics;   /* Number of graphics registers. */
    unsigned char  numAttribute;  /* Number of attribute registers. */
} VGARegisters;


/*
 * Display Mode
 */
typedef enum {
    VGAModePHSync      = 0x0001,
    VGAModeNHSync      = 0x0002,
    VGAModePVSync      = 0x0004,
    VGAModeNVSync      = 0x0008,
    VGAModeInterlace   = 0x0010,
    VGAModeDoubleScan  = 0x0020,
    VGAModeCSync       = 0x0040,
    VGAModePCSync      = 0x0080,
    VGAModeNCSync      = 0x0100,
    VGAModeHSkew       = 0x0200,  /* HSkew provided. */
    VGAModeBCast       = 0x0400,
    VGAModePixMux      = 0x1000,
    VGAModeDoubleClock = 0x2000,
    VGAModeClockDiv2   = 0x4000
} VGAModeFlags;

typedef struct {
    int clock;
    int clockIndex;
    
    unsigned int  depth;
    unsigned char biosMode;

    int hDisplay;
    int hBlankStart;
    int hBlankEnd;
    int hSyncStart;
    int hSyncEnd;
    int hTotal;
    int hSkew;
    int vDisplay;
    int vBlankStart;
    int vBlankEnd;
    int vSyncStart;
    int vSyncEnd;
    int vTotal;
    int vScan;
    int flags;
} VGAMode;


/*
 * Register Access Functions Type Definitions
 */
typedef struct _VGADescription *VGADescriptionRef;

typedef unsigned char (*VGAReadIndexFunction)(VGADescriptionRef VGA, unsigned char value);
typedef void (*VGAWriteIndexFunction)(VGADescriptionRef VGA, unsigned char index, unsigned char value);
typedef unsigned char (*VGAReadFunction)(VGADescriptionRef VGA);
typedef void (*VGAWriteFunction)(VGADescriptionRef VGA, unsigned char value);
typedef void (*VGAMiscFunction)(VGADescriptionRef VGA);


/*
 * VGA Controller Description
 */
typedef struct _VGADescription {
    IOEISAPortAddress IOBase;
    IOEISAPortAddress PIOOffset;
    
    unsigned char paletteEnabled;
    
    VGAReadIndexFunction  readCRTC;
    VGAWriteIndexFunction writeCRTC;
    VGAReadIndexFunction  readGraphics;
    VGAWriteIndexFunction writeGraphics;
    VGAReadIndexFunction  readAttribute;
    VGAWriteIndexFunction writeAttribute;
    VGAReadIndexFunction  readSequencer;
    VGAWriteIndexFunction writeSequencer;
    VGAReadFunction       readMiscOut;
    VGAWriteFunction      writeMiscOut;
    VGAMiscFunction       enablePalette;
    VGAMiscFunction       disablePalette;
    VGAWriteFunction      writeDACWriteAddress;
    VGAWriteFunction      writeDACData;
} VGADescription;

/*
 * Function Definitions
 */
unsigned char VGAGetIOBase(VGADescription *VGA);

void VGASequencerReset(VGADescription *VGA, unsigned char start);
void VGAProtect(VGADescription *VGA);
void VGAUnprotect(VGADescription *VGA);
void VGALock(VGADescription *VGA);
void VGAUnlock(VGADescription *VGA);

void VGARegistersAlloc(VGARegisters *registers);
void VGARegistersFree(VGARegisters *registers);
void VGAGetDescription(VGADescription *VGA, VGARegisters *registers);

void VGASave(VGADescription *VGA, VGARegisters *savedRegisters);
void VGARestore(VGADescription *VGA, VGARegisters *restoredRegisters);
void VGAEnterMode(VGADescription *VGA, VGARegisters *registers, const VGAMode *mode);

void VGARegistersLog(const VGARegisters *reg, const char *module);

#endif  /* VGA_H__ */
